//modified from briey
package vexriscv.demo

import vexriscv.plugin.{TightlyCoupledBus, _}
import vexriscv._
import spinal.core.Mem._
import spinal.core.sim.delayed
import vexriscv.ip.{DataCacheConfig, InstructionCacheConfig}
import spinal.core.{Bits, _}
import spinal.lib._
import spinal.lib.bus.amba3.apb._
import spinal.lib.bus.amba4.axi._
import spinal.lib.com.jtag.Jtag
import spinal.lib.com.jtag.sim.JtagTcp
import spinal.lib.com.uart.sim.{UartDecoder, UartEncoder}
import spinal.lib.com.uart.{Apb3UartCtrl, Uart, UartCtrlGenerics, UartCtrlMemoryMappedConfig}
import spinal.lib.graphic.RgbConfig
import spinal.lib.graphic.vga.{Axi4VgaCtrl, Axi4VgaCtrlGenerics, Vga}
import spinal.lib.io.{InOutWrapper, TriStateArray}
import spinal.lib.memory.sdram.SdramGeneration.SDR
import spinal.lib.memory.sdram._
import spinal.lib.memory.sdram.sdr.sim.SdramModel
import spinal.lib.memory.sdram.sdr.{Axi4SharedSdramCtrl, IS42x320D, MT48LC16M16A2, SdramInterface, SdramTimings}
import spinal.lib.misc.HexTools
import spinal.lib.soc.pinsec.{PinsecTimerCtrl, PinsecTimerCtrlExternal}
import spinal.lib.system.debugger.{JtagAxi4SharedDebugger, JtagBridge, SystemDebugger, SystemDebuggerConfig}
import spinal.lib.bus.misc.SizeMapping

import scala.collection.mutable.ArrayBuffer
import scala.collection.Seq

case class RyanConfig(axiFrequency : HertzNumber,
                      iTcmConfig : TcmConfig,
                      dTcmConfig : TcmConfig,
                      debunceTimeStep : TimeNumber,
                      sdramLayout: SdramLayout,
                      onChipRamSize : BigInt,
                      cpuPlugins : ArrayBuffer[Plugin[VexRiscv]],
                      uartCtrlConfig : UartCtrlMemoryMappedConfig)

object RyanConfig{
  def default = {
    val config = RyanConfig(
      axiFrequency = 81.25 MHz, //because of FPGA critical path @ iTCM, have to decrease from 100 to 81.25 MHz
      iTcmConfig = TcmConfig(),
      dTcmConfig = TcmConfig(),
      debunceTimeStep = 20 ms,
      sdramLayout = MT41K256M16TW.layout,
      onChipRamSize  = 8 KiB,
      uartCtrlConfig = UartCtrlMemoryMappedConfig(
        uartCtrlConfig = UartCtrlGenerics(
          dataWidthMax      = 8,
          clockDividerWidth = 20,
          preSamplingSize   = 1,
          samplingSize      = 5,
          postSamplingSize  = 2
        ),
        txFifoDepth = 16,
        rxFifoDepth = 16
      ),
      cpuPlugins = ArrayBuffer(
        //PcManagerSimplePlugin is not used anymore
        //new PcManagerSimplePlugin(0x80000000l, false),
        //          new IBusSimplePlugin(
        //            interfaceKeepData = false,
        //            catchAccessFault = true
        //          ),
        new IBusCachedPlugin(
          resetVector = 0x80000000l,
          //original
          //prediction = STATIC,
          //max perf
          prediction = DYNAMIC_TARGET,
          historyRamSizeLog2 = 8,
          config = InstructionCacheConfig(
            cacheSize = 2048,
            bytePerLine =32,
            wayCount = 1,
            addressWidth = 32,
            cpuDataWidth = 32,
            memDataWidth = 32,
            catchIllegalAccess = true,
            catchAccessFault = true,
            asyncTagMemory = false,
            //twoCycleRam = true,
            twoCycleRam = false,  //max perf
            twoCycleCache = true
          )
        ),
        new DBusCachedPlugin(
          config = new DataCacheConfig(
            cacheSize         = 2048,
            bytePerLine       = 32,
            wayCount          = 1,
            addressWidth      = 32,
            cpuDataWidth      = 32,
            memDataWidth      = 32,
            catchAccessError  = true,
            catchIllegal      = true,
            catchUnaligned    = true
          ),
          memoryTranslatorPortConfig = null   //這一項目應該是給linux用的TLB
          //            memoryTranslatorPortConfig = MemoryTranslatorPortConfig(
          //              portTlbSize = 6
          //            )
        ),
        new StaticMemoryTranslatorPlugin(
          ioRange      = _(31 downto 28) === 0xF
        ),
        new DecoderSimplePlugin(
          catchIllegalInstruction = true
        ),
        new RegFilePlugin(
          regFileReadyKind = plugin.SYNC,
          zeroBoot = false
        ),
        new IntAluPlugin,
        //SrcPlugin的功能是用來切換暫存器的來源
        new SrcPlugin(
          separatedAddSub = false,
          executeInsertion = true
        ),
        new FullBarrelShifterPlugin(earlyInjection = true),
        new MulPlugin,
        new DivPlugin,
        new HazardSimplePlugin(
          bypassExecute           = true,
          bypassMemory            = true,
          bypassWriteBack         = true,
          bypassWriteBackBuffer   = true,
          pessimisticUseSrc       = false,
          pessimisticWriteRegFile = false,
          pessimisticAddressMatch = false
        ),
        new BranchPlugin(
          earlyBranch = false,
          catchAddressMisaligned = true
        ),
        //simple interrupt mode
//        new CsrPlugin(
//          config = CsrPluginConfig(
//            catchIllegalAccess = false,
//            mvendorid      = null,
//            marchid        = null,
//            mimpid         = null,
//            mhartid        = null,
//            misaExtensionsInit = 66,
//            misaAccess     = CsrAccess.NONE,
//            mtvecAccess    = CsrAccess.NONE,
//            mtvecInit      = 0x80000020l,
//            mepcAccess     = CsrAccess.READ_WRITE,
//            mscratchGen    = false,
//            mcauseAccess   = CsrAccess.READ_ONLY,
//            mbadaddrAccess = CsrAccess.READ_ONLY,
//            mcycleAccess   = CsrAccess.NONE,
//            minstretAccess = CsrAccess.NONE,
//            ecallGen       = false,
//            wfiGenAsWait         = false,
//            ucycleAccess   = CsrAccess.NONE,
//            uinstretAccess = CsrAccess.NONE
//          )
//        ),
        // Vector interrupt mode
        new CsrPlugin(
          CsrPluginConfig.smallest.copy(
            xtvecModeGen = true,
            mtvecAccess = CsrAccess.WRITE_ONLY
          )
        ),
        new UserInterruptPlugin(
          interruptName = "vectorInt0",
          code = 20
        ),
//        new UserInterruptPlugin(
//          interruptName = "rawrrr",
//          code = 24
//        ),
        new YamlPlugin("cpu0.yaml")
      )
    )
    config
  }
}



class Ryan(val config: RyanConfig) extends Component{

  //Legacy constructor
  //Scala this关键字
   def this(axiFrequency: HertzNumber) {
    this(RyanConfig.default.copy(axiFrequency = axiFrequency))
  }

  import config._
  val debug = true
  //val interruptCount = 4
  //def vgaRgbConfig = RgbConfig(5,6,5)
  //ryan add
  val axi4config = Axi4Config(addressWidth=32, dataWidth=32, useId=false).toFullConfig()

  val io = new Bundle{
    //Clocks / reset
    //val asyncReset = in Bool()
    //val axiClk     = in Bool()
    val refClkIn = in Bool()
    val systemAsyncRstn    = in Bool()

    //Main components IO
    val jtag       = slave(Jtag())

    //Peripherals IO
    val gpioA         = master(TriStateArray(32 bits))
    //val gpioB         = master(TriStateArray(32 bits))
    val uart          = master(Uart())
    val timerExternal = in(PinsecTimerCtrlExternal())
    val coreInterruptn = in Bool()
    //val axi4Bus       = master(Axi4(axi4config))
    val ddr3IF        = Ddr3CtrlIF(config.sdramLayout)
  }
  noIoPrefix()

  val coreInterrupt = ~ io.coreInterruptn

  //⽤函数给⿊盒中所有信号重命名
  private def renameIO(): Unit = {
    io.flatten.foreach(bt => {
      if (bt.getName().contains("ddr3IF")) bt.setName(bt.getName().replace("ddr3IF_", ""))
    })
  }
  //在创建模块后执⾏重命名函数
  addPrePopTask(() => renameIO())

  val systemClkGen = new SystemClkGen()
  systemClkGen.io.clk_in1 <> io.refClkIn.asBits
  systemClkGen.io.resetn <> io.systemAsyncRstn.asBits

  //instantiate ddr3ctrl and gen clk and ret signal for cpu and bus
  val vivadoMigWrapper = new VivadoMigWrapper(config.sdramLayout)
  vivadoMigWrapper.io.systemClk <> systemClkGen.io.clk_out1.asBool //100MHz
  vivadoMigWrapper.io.systemResetn <> systemClkGen.io.locked.asBool

  val fromDebug = Bool()
  fromDebug := False
  val axiClk = vivadoMigWrapper.io.busClk
  val cpuReset = vivadoMigWrapper.io.cpuReset
  val axiReset = vivadoMigWrapper.io.busReset || fromDebug || cpuReset
  val debugReset = vivadoMigWrapper.io.busReset || cpuReset
  io.ddr3IF <> vivadoMigWrapper.io.ddr3IF


//  val resetCtrlClockDomain = ClockDomain(
//    clock = io.axiClk,
//    config = ClockDomainConfig(
//      resetKind = BOOT
//    )
//  )
//
//  val resetCtrl = new ClockingArea(resetCtrlClockDomain) {
//    val systemResetUnbuffered  = False
//    //    val coreResetUnbuffered = False
//
//    //Implement an counter to keep the reset axiResetOrder high 64 cycles
//    // Also this counter will automaticly do a reset when the system boot.
//    val systemResetCounter = Reg(UInt(6 bits)) init(0)
//    when(systemResetCounter =/= U(systemResetCounter.range -> true)){
//      systemResetCounter := systemResetCounter + 1
//      systemResetUnbuffered := True
//    }
//    when(BufferCC(io.asyncReset)){
//      systemResetCounter := 0
//    }
//
//    //Create all reset used later in the design
//    val systemReset  = RegNext(systemResetUnbuffered)
//    val axiReset     = RegNext(systemResetUnbuffered)
//  }

//  val axiClockDomain = ClockDomain(
//    clock = io.axiClk,
//    reset = resetCtrl.axiReset,
//    frequency = FixedFrequency(axiFrequency) //The frequency information is used by the SDRAM controller
//  )
  val axiClockDomain = ClockDomain(
    clock = axiClk,
    reset = axiReset,
    frequency = FixedFrequency(axiFrequency)
  )

//  val debugClockDomain = ClockDomain(
//    clock = io.axiClk,
//    reset = resetCtrl.systemReset,
//    frequency = FixedFrequency(axiFrequency)
//  )
  val debugClockDomain = ClockDomain(
    clock = axiClk,
    reset = debugReset,
    frequency = FixedFrequency(axiFrequency)
  )

  val axi = new ClockingArea(axiClockDomain) {
    val ram = Axi4SharedOnChipRam(
      dataWidth = 32,
      byteCount = onChipRamSize,
      idWidth = 4
    )

    val apbBridge = Axi4SharedToApb3Bridge(
      addressWidth = 20,
      dataWidth    = 32,
      idWidth      = 4
    )

//    val gpioACtrl = Apb3Gpio(
//      gpioWidth = 32,
//      withReadSync = true
//    )
    val gpioACtrl = Apb3GpioIntCtrl(
      gpioWidth = 32,
      withReadSync = true,
      debunceTimeStep = config.debunceTimeStep
    )
//    val gpioBCtrl = Apb3Gpio(
//      gpioWidth = 32,
//      withReadSync = true
//    )
    val timerCtrl = PinsecTimerCtrl()

//    var temp = Seq(gpioACtrl.io.gpio.read(31 downto 30).setName("gpioRead"),
//      gpioACtrl.io.gpio.writeEnable(31 downto 30).setName("writeEnable"),
//      gpioACtrl.intEnable.pull()(31 downto 30).setName("intEnable"),
//      gpioACtrl.intPolarity.pull()(31 downto 30).setName("intPolarity"),
//      gpioACtrl.io.intSignal(31 downto 30).setName("intSignal") )
//    val ilaGpio = Ila(temp,moduleName = "gpio")

    val uartCtrl = Apb3UartCtrl(uartCtrlConfig)
    uartCtrl.io.apb.addAttribute(Verilator.public)

    val core = new Area{
      cpuPlugins += new DebugPlugin(debugClockDomain)
      if (iTcmConfig.tcmUsed == true) {
        cpuPlugins += new IBusDBusCachedTightlyCoupledRam(
          mapping = SizeMapping(iTcmConfig.tcmRamAddress, iTcmConfig.tcmRamSize),
          ramAsBlackbox = iTcmConfig.ramAsBlackbox,
          hexInit = iTcmConfig.hexInit,
          ramOffset = iTcmConfig.tcmRamAddress.toLong
        )
        //If iTcm exists, modify the resetVector
        for(plugin <- cpuPlugins) plugin match{
          case plugin: IBusCachedPlugin => plugin.resetVector = iTcmConfig.tcmRamAddress
          //case plugin: IBusCachedPlugin => println("test")
          case _ =>
        }
      }
      if (dTcmConfig.tcmUsed == true) {
        cpuPlugins += new IBusDBusCachedTightlyCoupledRam(
          mapping = SizeMapping(dTcmConfig.tcmRamAddress, dTcmConfig.tcmRamSize),
          ramAsBlackbox = dTcmConfig.ramAsBlackbox,
          hexInit = dTcmConfig.hexInit,
          ramOffset = dTcmConfig.tcmRamAddress.toLong
        )
      }
      val config = VexRiscvConfig(plugins = cpuPlugins)

      val cpu = new VexRiscv(config)
      var iBus: Axi4ReadOnly = null
      var dBus: Axi4Shared = null
      for(plugin <- config.plugins) plugin match{
        //case plugin : IBusSimplePlugin => iBus = plugin.iBus.toAxi4ReadOnly()
//        case plugin: IBusDBusCachedTightlyCoupledRam => {
//          println(plugin.tcmMemIF.addressWidth)
//          var tcmMemIF = iTcmMemIF
//          if (plugin.tcmMemIF.addressWidth == 11) tcmMemIF=iTcmMemIF else tcmMemIF=dTcmMemIF
//          plugin.tcmMemIF.readData := tcmMemIF.readData
//          tcmMemIF.address := plugin.tcmMemIF.address
//          tcmMemIF.enable := plugin.tcmMemIF.enable
//          tcmMemIF.writeData := plugin.tcmMemIF.writeData
//          tcmMemIF.write := plugin.tcmMemIF.write
//          tcmMemIF.mask := plugin.tcmMemIF.mask
//        }
        case plugin : IBusCachedPlugin => {
          if (iTcmConfig.tcmUsed == true){
            plugin.resetVector = iTcmConfig.tcmRamAddress
          }
          iBus = plugin.iBus.toAxi4ReadOnly()
        }
        //case plugin : DBusSimplePlugin => dBus = plugin.dBus.toAxi4Shared()
        case plugin : DBusCachedPlugin => dBus = plugin.dBus.toAxi4Shared(true)
        case plugin : CsrPlugin        => {
          plugin.externalInterrupt := BufferCC(coreInterrupt)
          plugin.timerInterrupt := timerCtrl.io.interrupt
        }
        case plugin : UserInterruptPlugin => {
          plugin.interrupt := gpioACtrl.io.intSignal(31)
          //plugin.interrupt := vectoredIntDebounce(0)
          //plugin.interrupt := vectoredIntDebounce(0).rise()
        }
        case plugin : DebugPlugin      => debugClockDomain{
          //axiReset setWhen(RegNext(plugin.io.resetOut))
          when(RegNext(plugin.io.resetOut)){fromDebug:=True}
          io.jtag <> plugin.io.bus.fromJtag()
        }
        case _ =>
      }
    }

    val axi4SharedToAxi4 = new Area {
      val sAxi4Shared = Axi4Shared(axi4config)
      val mAxi4 = sAxi4Shared.toAxi4().toFullConfig()
    }

    val axiCrossbar = Axi4CrossbarFactory()

    axiCrossbar.addSlaves(
      //use -> to combine tuple
      ram.io.axi                    -> (0x80000000L,   onChipRamSize),
      axi4SharedToAxi4.sAxi4Shared  -> (0x40000000L,   512 MiB),
      apbBridge.io.axi              -> (0xF0000000L,   1 MiB)
    )

    axiCrossbar.addConnections(
      core.iBus       -> List(ram.io.axi, axi4SharedToAxi4.sAxi4Shared ),
      core.dBus       -> List(ram.io.axi, axi4SharedToAxi4.sAxi4Shared  , apbBridge.io.axi)
    )


    axiCrossbar.addPipelining(apbBridge.io.axi)((crossbar,bridge) => {
      crossbar.sharedCmd.halfPipe() >> bridge.sharedCmd
      crossbar.writeData.halfPipe() >> bridge.writeData
      crossbar.writeRsp             << bridge.writeRsp
      crossbar.readRsp              << bridge.readRsp
    })

    axiCrossbar.addPipelining(axi4SharedToAxi4.sAxi4Shared)((crossbar,ctrl) => {
      crossbar.sharedCmd.halfPipe()  >>  ctrl.sharedCmd
      crossbar.writeData            >/-> ctrl.writeData
      crossbar.writeRsp              <<  ctrl.writeRsp
      crossbar.readRsp               <<  ctrl.readRsp
    })

    axiCrossbar.addPipelining(ram.io.axi)((crossbar,ctrl) => {
      crossbar.sharedCmd.halfPipe()  >>  ctrl.sharedCmd
      crossbar.writeData            >/-> ctrl.writeData
      crossbar.writeRsp              <<  ctrl.writeRsp
      crossbar.readRsp               <<  ctrl.readRsp
    })

    axiCrossbar.addPipelining(core.dBus)((cpu,crossbar) => {
      cpu.sharedCmd             >>  crossbar.sharedCmd
      cpu.writeData             >>  crossbar.writeData
      cpu.writeRsp              <<  crossbar.writeRsp
      cpu.readRsp               <-< crossbar.readRsp //Data cache directly use read responses without buffering, so pipeline it for FMax
    })

    axiCrossbar.build()


    val apbDecoder = Apb3Decoder(
      master = apbBridge.io.apb,
      slaves = List(
        gpioACtrl.io.apb -> (0x00000, 4 KiB),
        //gpioBCtrl.io.apb -> (0x01000, 4 KiB),
        uartCtrl.io.apb  -> (0x10000, 4 KiB),
        timerCtrl.io.apb -> (0x20000, 4 KiB)
      )
    )
  }

  io.gpioA                      <> axi.gpioACtrl.io.gpio
  //io.gpioB                      <> axi.gpioBCtrl.io.gpio
  io.timerExternal              <> axi.timerCtrl.io.external
  io.uart                       <> axi.uartCtrl.io.uart
  vivadoMigWrapper.io.s_axi4Bus <> axi.axi4SharedToAxi4.mAxi4
}

//DE1-SoC
object Ryan{
  def main(args: Array[String]) {
    val config = SpinalConfig()
    val iTcmConfig = TcmConfig(tcmUsed = false, tcmRamAddress = 0x20000000, tcmRamSize = 0x2000)
    val dTcmConfig = TcmConfig(tcmUsed = false, tcmRamAddress = 0x30000000, tcmRamSize = 0x1000)

    config.generateVerilog({
      val toplevel =  new Ryan(RyanConfig.default.copy(iTcmConfig=iTcmConfig,dTcmConfig=dTcmConfig))
      InOutWrapper(toplevel)
    })
  }
}

//DE1-SoC with memory init
object RyanWithMemoryInit{
  def main(args: Array[String]) {
    val config = SpinalConfig()
    //for TCM simulation
    if (true){
      val config = SpinalConfig()
      val iTcmConfig = TcmConfig(tcmUsed = true, ramAsBlackbox = false, tcmRamAddress = 0x20000000, tcmRamSize = 0x4000, hexInit = "/home/datakey/myProject/VexRiscvSocSoftware-master/projects/briey/timer/build/timerITCM.hex")
      val dTcmConfig = TcmConfig(tcmUsed = true, ramAsBlackbox = false, tcmRamAddress = 0x30000000, tcmRamSize = 0x4000, hexInit = "/home/datakey/myProject/VexRiscvSocSoftware-master/projects/briey/timer/build/timerDTCM.hex")
//      val iTcmConfig = TcmConfig(tcmUsed = true, ramAsBlackbox = false, tcmRamAddress = 0x20000000, tcmRamSize = 0x2000, hexInit = "/home/datakey/myProject/VexRiscvSocSoftware-master/projects/briey/timer/build/timerITCM.hex")
//      val dTcmConfig = TcmConfig(tcmUsed = true, ramAsBlackbox = false, tcmRamAddress = 0x30000000, tcmRamSize = 0x1000, hexInit = "/home/datakey/myProject/VexRiscvSocSoftware-master/projects/briey/timer/build/timerDTCM.hex")

      config.generateVerilog({
        val toplevel = new Ryan(RyanConfig.default.copy(iTcmConfig = iTcmConfig, dTcmConfig = dTcmConfig))
        HexTools.initRam(toplevel.axi.ram.ram, "src/main/ressource/hex/muraxDemo.hex", 0x80000000l)
        InOutWrapper(toplevel)
        }
      )
    }
  }
}

//set debunceTimeStep only 100ns to short simulation time
object RyanWithMemoryInitForVerilogSim{
  def main(args: Array[String]) {
    val config = SpinalConfig()
    //for TCM simulation
    if (true){
      val config = SpinalConfig()
      val iTcmConfig = TcmConfig(tcmUsed = true, ramAsBlackbox = false, tcmRamAddress = 0x20000000, tcmRamSize = 0x2000, hexInit = "/home/datakey/myProject/VexRiscvSocSoftware-master/projects/briey/timer/build/timerITCM.hex")
      val dTcmConfig = TcmConfig(tcmUsed = true, ramAsBlackbox = false, tcmRamAddress = 0x30000000, tcmRamSize = 0x1000, hexInit = "/home/datakey/myProject/VexRiscvSocSoftware-master/projects/briey/timer/build/timerDTCM.hex")

      config.generateVerilog({
        val toplevel = new Ryan(RyanConfig.default.copy(iTcmConfig = iTcmConfig, dTcmConfig = dTcmConfig, debunceTimeStep = 100 ns))
        HexTools.initRam(toplevel.axi.ram.ram, "src/main/ressource/hex/muraxDemo.hex", 0x80000000l)
        InOutWrapper(toplevel)
      }
      )
    }
  }
}


import spinal.core.sim._
object RyanSim {
  def main(args: Array[String]): Unit = {
    val simSlowDown = false
    SimConfig.allOptimisation.compile(new Ryan(RyanConfig.default)).doSimUntilVoid{dut =>
      val mainClkPeriod = (1e12/dut.config.axiFrequency.toDouble).toLong
      val jtagClkPeriod = mainClkPeriod*4
      val uartBaudRate = 115200
      val uartBaudPeriod = (1e12/uartBaudRate).toLong

      val clockDomain = ClockDomain(dut.axiClk, dut.axiReset)
      clockDomain.forkStimulus(mainClkPeriod)

      val tcpJtag = JtagTcp(
        jtag = dut.io.jtag,
        jtagClkPeriod = jtagClkPeriod
      )

      val uartTx = UartDecoder(
        uartPin = dut.io.uart.txd,
        baudPeriod = uartBaudPeriod
      )

      val uartRx = UartEncoder(
        uartPin = dut.io.uart.rxd,
        baudPeriod = uartBaudPeriod
      )


      dut.io.coreInterruptn #= true
    }
  }
}
